/* ************************************************************************** */
/*                                       _____           _                    */
/*  packet/recv.c                       |_   _|__  _   _| |__   ___ _   _     */
/*  | Project: libp7                      | |/ _ \| | | | '_ \ / _ \ | | |    */
/*                                        | | (_) | |_| | | | |  __/ |_| |    */
/*  By: thomas <thomas@touhey.fr>         |_|\___/ \__,_|_| |_|\___|\__, |.fr */
/*  Last updated: 2017/01/04 15:01:48                               |___/     */
/*                                                                            */
/* ************************************************************************** */
#include <libp7/internals.h>
#include <stdlib.h>
#include <string.h>

/**
 *	complete_packet:
 *	Complete the packet with N bytes.
 *
 *	Uses stream buffering to achieve this.
 *
 *	@arg	handle		the libp7 handle.
 *	@arg	buf			the buffer.
 *	@arg	recv		pointer to currently received bytes.
 *	@arg	with		complete with this number of byte.
 *	@return				the error (0 if ok)
 */

static int complete_packet(p7_handle_t *handle,
	unsigned char *buf, size_t *recv, size_t with)
{
	int err = p7_read(handle->_stream, &buf[*recv], with, ACTIVE_TIMEOUT);
	*recv += with;
	return (err);
}

/**
 *	p7_recv_main:
 *	Receive a packet.
 *
 *	This is the main receiving function. It receives, parses, and returns
 *	if everything has been successfully done or not.
 *	It uses buffered I/O to read progressively all of the packet.
 *
 *	@arg	handle		the libp7 handler
 *	@return				the error (0 if ok)
 */

#define buffer handle->_recv_buffer
#define COMPLETE_PACKET(N) { \
	int COMP_PACKET_err; \
	if ((COMP_PACKET_err = complete_packet(handle, \
	  buffer, &received, (N)))) \
		return (COMP_PACKET_err); \
}

static int p7_recv_main(p7_handle_t *handle)
{
	/* prepare reception */
	size_t received = 0;

	/* log packet */
	log_info("receiving packet...");

	/* check type */
	COMPLETE_PACKET(1)
	response.type = buffer[0];

	/* image has a particular format starting from here, look for it now! */
	if (response.type == p7_pt_ohp) {
		COMPLETE_PACKET(5)
		/* get image by its type */
		unsigned int image_size = 0;
		if (!memcmp(&buffer[1], "TYP01", 5)) {
			log_info("This is normal VRAM, we know it !");
			response.scrtype = p7_pscr_typ01;
			image_size = 1024;
		} else {
			log_error("Unknown image type : %.5s", &buffer[1]);
			return (p7_error_checksum);
		}

		/* complete packet */
		log_info("Get screen and checksum (%uo)", image_size + 2);
		COMPLETE_PACKET(image_size + 2)
		memcpy(&response.vram, &buffer[6], image_size);

		/* log */
		unsigned int packet_size = 6 + image_size + 2;
		log_info("received the following [screen] packet (%uo) :", packet_size);
		logm_info(buffer, packet_size);

		/* check the sum */
		if (!image_size
		 || p7_checksum(buffer, packet_size)
		 != p7_getascii(&buffer[packet_size - 2], 2))
			return (p7_error_checksum);

		/* and return the packet */
		return (0);
	}

	/* subtype and extended */
	COMPLETE_PACKET(3)
	p7ushort_t subtype = p7_getascii(&buffer[1], 2);

	/* - extended (beginning) - */
	p7ushort_t data_size = 0;
	int is_extended = (buffer[3] == '1');
	if (is_extended) {
		/* get data size */
		COMPLETE_PACKET(4)
		data_size = p7_getascii(&buffer[4], 4);

		/* get data */
		COMPLETE_PACKET(data_size + 2)
	} else
		COMPLETE_PACKET(2)

	/* check if we should read a binary zero */
	int is_binary_zero = (response.type == p7_pt_cmd
		&& subtype == 0x56);
	if (is_binary_zero) COMPLETE_PACKET(1)

	/* log */
	p7ushort_t packet_size = 6 + 4 * !!is_extended + data_size;
	log_info("received the following [normal] packet (%" PRIuP7SHORT "o) :",
		packet_size);
	logm_info(buffer, packet_size);

	/* calculate checksum */
	unsigned int calculated_checksum = p7_checksum(buffer, packet_size);
	unsigned int found_checksum = p7_getascii(&buffer[packet_size - 2], 2);
	if (calculated_checksum != found_checksum)
		return (p7_error_checksum);

	/* - extended (finish) - */
	if (is_extended)
		data_size = p7_decode(&buffer[8], &buffer[8], data_size);

	/* get fields out for specific packets */
	switch (response.type) {
	/* - for command - */
	case p7_pt_cmd:
		log_info("packet was interpreted as a command one");
		response.code = subtype;
		log_info("command is '%s' (0x%02" PRIxP7SHORT ")",
			p7_getcmdstring(subtype), subtype);
		if (is_extended
		 && p7_decode_command(handle, &buffer[8], data_size))
		 	return (p7_error_checksum);
		break;

	/* - for data - */
	case p7_pt_data:
		log_info("packet was interpreted as a data one");
		p7_decode_data(handle, &buffer[8], data_size);
		break;

	/* - for roleswap - */
	case p7_pt_roleswp:
		log_info("packet was interpreted as a roleswap one");
		log_info("becoming active again");
		handle->_active = 1;
		break;

	/* - for check - */
	case p7_pt_check:
		log_info("packet was interpreted as a check one");
		response.initial = !subtype;
		break;

	/* - for ack - */
	case p7_pt_ack:
		log_info("packet was interpreted as an ack one");
		response.ow = (subtype == 0x01);
		response.extended = (subtype == 0x02);
		if (response.extended
		 && p7_decode_ack(handle, &buffer[8], data_size))
			return (p7_error_checksum);
		if (subtype == 0x03)
			handle->_terminated = 1;
		break;

	/* - for error - */
	case p7_pt_error:
		log_info("packet was interpreted as an error one");
		log_info("error is '%s' (0x%02" PRIxP7SHORT ")",
			p7_geterrstring(subtype), subtype);
		response.code = subtype;
		if (subtype == p7_err_fullmem)
			handle->_terminated = 1;
		break;

	/* - for termination - */
	case p7_pt_terminate:
		log_info("packet was interpreted as a terminate one");
		log_info("termination cause is '%s' (0x%02" PRIxP7SHORT ")",
			p7_gettermstring(subtype), subtype);
		response.code = subtype;
		if (subtype == p7_term_user)
			return (p7_error_interrupted);
		break;

	/* other */
	default:
		log_fatal("bad type received ! type was 0x%02x", response.type);
		return (p7_error_unknown);
	}

	/* finally, return the packet */
	return (0);
}

/**
 *	p7_recv:
 *	Receives packet, checks for errors.
 *
 *	@arg	handle			the libp7 handler
 *	@arg	checksum		if 0 and invalid checksum, ignore and receive again
 *	@return					if it worked
 */

int p7_recv(p7_handle_t *handle, int checksum)
{
	/* check if handler is initialized */
	if (!handle)
		return (p7_error_uninitialized);

	/* main receiving loop */
	int tries = 2, err = 0;
	int wasresend = 0;
	do {
		/* get packet */
		err = p7_recv_main(handle);

		/* if there was a fail */
		if (err) switch (err) {
			/* if calc couldn't be found, terminate communication */
			case p7_error_nocalc:
				tries = 0;
				break;

			/* if it is a timeout, remove a try */
			case p7_error_timeout:
				tries--;

				/* send a check */
				err = p7_send_check(handle);
				if (err) return (err);

				/* set things */
				wasresend = 0;
				err = p7_error_timeout;
				break;

			/* if it is a bad checksum, ask for resend ; also,
			 * reset tries number in case of bad checksum between timeouts */
			case p7_error_checksum:
				/* if we shouldn't bother sending resend error, just
				 * transmit checksum error */
				if (!checksum) break;

				/* if we receive an invalid checksum while we are in packet
				 * shifting mode, or if we received an error when we sent a
				 * resend error, cut connexion... we can't do anything. */
				if (handle->_shifted || wasresend)
					return (p7_error_irrecoverable);

				/* otherwise, send resend error */
				err = p7_send_err_resend(handle);
				if (err) return (err);

				/* set things */
				wasresend = 1;
				err = p7_error_checksum;
				break;

			/* should not catch bad type & others */
			default:
				return (p7_error_unknown);
		}
	} while (err && tries);

	/* return the error */
	return (err);
}
