/* ************************************************************************** */
/*                                       _____           _                    */
/*  protocol/request.c                  |_   _|__  _   _| |__   ___ _   _     */
/*  | Project: libp7                      | |/ _ \| | | | '_ \ / _ \ | | |    */
/*                                        | | (_) | |_| | | | |  __/ |_| |    */
/*  By: thomas <thomas@touhey.fr>         |_|\___/ \__,_|_| |_|\___|\__, |.fr */
/*  Last updated: 2017/01/04 15:01:48                               |___/     */
/*                                                                            */
/* ************************************************************************** */
#include <libp7/internals.h>

/**
 *	p7_request:
 *	Requests a file.
 *
 *	@arg	handle		the libp7 handle
 *	@arg	buffer		the buffer (should be writable)
 *	@arg	dirname		the directory name
 *	@arg	filename	the filename
 *	@arg	devname		the device name
 *	@arg	disp		callback display
 *	@return				if it worked
 */

int p7_request(p7_handle_t *handle, const p7_buffer_t *buffer,
	const char *dirname, const char *filename, const char *devname,
	void (*disp)(p7ushort_t, p7ushort_t))
{
	int err;

	/* make checks */
	chk_required_filename(filename);
	chk_dirname(dirname);
	chk_bufwrite(buffer);
	chk_handle(handle);
	chk_active(handle);

	/* send command packet */
	log_info("sending file transfer request");
	if ((err = p7_send_cmdfls_reqfile(handle, dirname, filename, devname))) {
		log_fatal("couldn't send file transfer request/didn't receive answer");
		return (err);
	}

	/* check response packet */
	if (response.type == p7_pt_error
	 && response.code == p7_err_other) {
		log_fatal("requested file doesn't exist");
		return (p7_error_notfound);
	} else if (response.type != p7_pt_ack) {
		log_fatal("didn't receive ack or known error...");
		return (p7_error_unknown);
	}

	/* swap roles */
	log_info("sending roleswap");
	if ((err = p7_send_roleswp(handle))) {
		log_fatal("couldn't swap roles");
		return (err);
	} else if (response.type != p7_pt_cmd || response.code != 0x45) {
		log_fatal("didn't receive expected command");
		return (p7_error_unknown);
	}

	/* get data */
	if ((err = p7_get_buffer(handle, buffer, response.filesize, 0, disp)))
		return (err);

	/* check if last answer was roleswap */
	if (response.type != p7_pt_roleswp)
		return (p7_error_unknown);

	/* no error! */
	return (0);
}

#ifndef P7_DISABLED_FILE
/**
 *	p7_reqfile:
 *	Request a FILE.
 *
 *	@arg	handle		the handle.
 *	@arg	file		the FILE.
 *	@arg	dirname		the directory name
 *	@arg	filename	the file name
 *	@arg	devname		the storage device name
 *	@arg	disp		the display callback
 *	@return				the error code (0 if ok).
 */

int p7_reqfile(p7_handle_t *handle, FILE *file,
	const char *dirname, const char *filename, const char *devname,
	void (*disp)(p7ushort_t, p7ushort_t))
{
	chk_iswrite(file);
	p7_buffer_t buffer = {
		.cookie = file,
		.write = p7_filebuffer_write
	};

	return (p7_request(handle, &buffer, dirname, filename, devname, disp));
}

#endif
