/* ************************************************************************** */
/*                                       _____           _                    */
/*  protocol/sendexe.c                  |_   _|__  _   _| |__   ___ _   _     */
/*  | Project: libp7                      | |/ _ \| | | | '_ \ / _ \ | | |    */
/*                                        | | (_) | |_| | | | |  __/ |_| |    */
/*  By: thomas <thomas@touhey.fr>         |_|\___/ \__,_|_| |_|\___|\__, |.fr */
/*  Last updated: 2017/01/04 15:01:48                               |___/     */
/*                                                                            */
/* ************************************************************************** */
#include <libp7/internals.h>

/**
 *	p7_sendexe:
 *	Send executable stream.
 *
 *	@arg	handle		the libp7 handle
 *	@arg	buffer		the update.exe to send.
 *	@arg	loadaddr	the load address
 *	@arg	straddr		the start address
 *	@arg	disp		the display callback
 *	@return				0 if it worked, error code otherwise
 */

int p7_sendexe(p7_handle_t *handle, const p7_buffer_t *buffer,
	p7uint_t loadaddr, p7uint_t straddr,
	void (*disp)(p7ushort_t, p7ushort_t))
{
	int err;

	/* make checks */
	chk_bufread(buffer);
	chk_handle(handle);
	chk_active(handle);
	chk_filesize(buffer->size);

	/* send command */
	log_info("sending up and run command");
	if ((err = p7_send_cmdosu_upandrun(handle, buffer->size,
	  loadaddr, straddr)))
		return (err);
	else if (response.type == p7_pt_error && response.code == p7_err_other) {
		log_fatal("0x56 is non implemented!");
		return (p7_error_unsupported);
	} else if (response.type != p7_pt_ack) {
		log_fatal("response wasn't ack");
		return (p7_error_unknown);
	}

	/* upload executable */
	if ((err = p7_send_buffer(handle, buffer, 0, disp)))
		return (err);

	/* we're good */
	return (0);
}

#ifndef P7_DISABLED_FILE
/**
 *	p7_sendexe_file:
 *	Send executable.
 *
 *	@arg	handle		the libp7 handle
 *	@arg	exe			the update.exe to send
 *	@arg	loadaddr	the load address
 *	@arg	straddr		the start address
 *	@arg	disp		the display callback
 *	@return				if it worked
 */

int p7_sendexe_file(p7_handle_t *handle,
	FILE *exe, p7uint_t loadaddr, p7uint_t straddr,
	void (*disp)(p7ushort_t, p7ushort_t))
{
	/* make checks */
	chk_isread(exe);

	/* calculate exe size */
	p7uint_t size;
	if (fseek(exe, 0, SEEK_END)) {
		log_fatal("seek failed!");
		return (p7_error_noseek);
	}
	size = (p7uint_t)ftell(exe);
	rewind(exe);

	/* make the buffer */
	p7_buffer_t buffer = {
		.cookie = exe,
		.size = size,
		.read = p7_filebuffer_read
	};

	/* real func */
	return (p7_sendexe(handle, &buffer, loadaddr, straddr, disp));
}
#endif
